# python3-opsi
# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
# and published under the Terms of the GNU Affero General Public License.
# Text of the AGPL: http://www.gnu.org/licenses/agpl-3.0-standalone.html
# credits: http://www.opsi.org/credits/

#***************************************************************************
# Subversion:
# $Revision: 40 $
# $Author: oertel $
# $Date: 2015-01-16 18:17:00 +0100 (Fr, 16 Jan 2015) $
#***************************************************************************

# Debian
# https://www.debian.org/releases/stable/amd64/ch05s03.html.de
# https://wiki.debian.org/DebianInstaller/
# https://www.debian.org/releases/stable/amd64/apb.html.de
# preseed files:
# http://www.debian-administration.org/article/394/Automating_new_Debian_installations_with_preseeding

# Ubuntu:
#http://archive.ubuntu.com/ubuntu/dists/trusty-updates/main/installer-amd64/current/images/netboot/ubuntu-installer/amd64/
#https://wiki.ubuntu.com/UbiquityAutomation
#https://help.ubuntu.com/14.04/installation-guide/example-preseed.txt

try:
    import ConfigParser
except:
    import configparser
from OPSI.Util.File import IniFile
from OPSI.Util.File import TextFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from OPSI.Util import blowfishDecrypt
from hashlib import md5
try:
    from twisted.conch.ssh import keys
except:
    pass
try:
    from urlparse import urlparse
except:
    from urllib.parse import urlparse
from os.path import splitext, basename
import platform
import re, xml.dom.minidom, os, gzip
import crypt
import subprocess
from base64 import b64encode
from os import urandom

if not bootimageVersion or (int(bootimageVersion) < 20210618):
    raise Exception(u"This product requires opsi-linux-bootimage version >= 20210618")
try:
    if not backend.backend_info()['opsiVersion'].startswith('4.'):
        raise ValueError('No opsi 4.x backend present')
except Exception:
    raise Exception(u"This is a opsi 4.x product")

isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

if isOpsi43:
    show_message(u"Installing Product: %s" % (productId), 10)

opsiVersion = backend.backend_info()['opsiVersion']
source = SCRIPT_PATH

# Do hardware inventory
logger.notice(u"Fetching opsi hw audit configuration")
hwconfig = backend.auditHardware_getConfig()

if isOpsi43:
    show_message(u"Running hardware inventory", 10)
logger.notice(u"Running hardware inventory")
auditHardwareOnHosts = auditHardware(config = hwconfig, hostId = clientId)

logger.notice(u"Sending hardware information to service")
backend.auditHardwareOnHost_setObsolete(clientId)
backend.auditHardwareOnHost_createObjects(auditHardwareOnHosts)

ramSize = execute("grep MemTotal /proc/meminfo | awk '{ print $2 }'")[0]
if int(ramSize) < 5000000:
    raise Exception("Memory is less than 5GB. A network installation requires at least 5GB of RAM")

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')
exec(open('/tmp/opsisetuplib.py').read())

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsipreparelib.py','/tmp')
exec(open('/tmp/opsipreparelib.py').read())

initOpsiSetupLib()

useGpt = False
inUefi = inUefiMode()
inSecureboot = False
if inUefi:
    try:
        inSecureboot = isSecureboot()
    except NameError:
        isSecureboot = False

    if inSecureboot:
        logger.notice(u"we are running in Secure Boot mode")
        scriptMessageSubject.setMessage(u"we are running in Secure Boot mode")
    else:
        logger.notice(u"we are not running in Secure Boot mode")
        scriptMessageSubject.setMessage(u"we are not running in Secure Boot mode")
    useGpt = True

if useGpt:
    logger.notice(u"Will use GPT")
else:
    logger.notice(u"Will use MBR Partitions")

# create network path
if (depot._url.startswith('smb:') or depot._url.startswith('cifs:')):
    match = re.search('^smb://([^/]+)/([^/]+)(.*)$', depot._url, re.IGNORECASE)
    if not match:
        raise Exception("Bad depot-URL '%s'" % depot._url)
    hn = match.group(1)
    sn = match.group(2)
    pn = match.group(3)
else:
    match = re.search('^webdavs://([^/]+):([^/]+)(.*)$', depot._url, re.IGNORECASE)
    if not match:
        raise Exception("Bad depot-URL '%s'" % depot._url)
    hn = match.group(1)
    sn = match.group(2)
    pn = match.group(3)

depotDir = '/var/lib/opsi/depot'
nfsserver = hn
nfssubdir = 'opsi_nfs_share'
nfsbasedir = '%s/%s' % (depotDir, nfssubdir)
nfsclientconfig =  configStates.get('clientconfig.opsi_nfs_share', [ '' ])[0]
if nfsclientconfig:
    nfsserver = nfsclientconfig.split(":")[0]
    nfsbasedir = nfsclientconfig.split(":")[1]
logger.notice(u"Using as nfsserver: %s" % nfsserver)
logger.notice(u"Using as nfsbasedir: %s" % nfsbasedir)

proxy = productProperties.get('proxy','')

# Get and encrypt passwords
# user
username_from_property = ""
username_from_property = productProperties.get('user_name', 'user')
if not username_from_property:
    username = 'user'
else:
    username = username_from_property
userpassclear = productProperties.get('user_password', 'linux123')
userpassenc = shadow_password_hash(userpassclear)

# root
rootpassclear = productProperties.get('root_password', 'linux123')
rootpassenc = shadow_password_hash(rootpassclear)

desktop_package = productProperties.get('desktop_package', 'standard')
encrypt_password = productProperties.get('encrypt_password', '')

install_opsi_client_agent = False
do_opsi_client = "false"
opsi_client_agent_installation = '    - echo'
if (str(productProperties.get('install_opsi-client-agent', 'true')).lower() == 'true'):
    install_opsi_client_agent = True
    do_opsi_client = "true"
    opsi_client_agent_installation = '    - curtin in-target --target=/target -- apt install -y wget openssh-server lsb-release cifs-utils vim lsof dnsutils librsync-dev pkg-config\n\
    - curtin in-target --target=/target -- mkdir -p /var/tmp/opsi-linux-client-agent\n\
    - curtin in-target --target=/target -- wget -r -nH --no-check-certificate --cut-dirs=2 --no-parent --http-user #@client_id*# --http-passwd #@client_key*# --reject="index.html*" https://#@depotAddress*#:4447/depot/opsi-linux-client-agent/ -P /var/tmp/opsi-linux-client-agent\n\
    - curtin in-target --target=/target -- wget -r -nH --no-check-certificate --cut-dirs=3 --no-parent --http-user #@client_id*# --http-passwd #@client_key*# --reject="index.html*" https://#@depotAddress*#:4447/depot/#@productid*#/opsi/postinst.d/ -P /var/tmp/opsi-linux-client-agent/files\n\
    - curtin in-target --target=/target -- find /var/tmp/opsi-linux-client-agent/files/postinst.d -name "*.sh" -exec chmod -x {} +\n\
    - curtin in-target --target=/target -- chmod -R 777 /var/tmp/opsi-linux-client-agent/\n\
    - echo "CLIENT_ID=#@client_id*#" >> /target/var/tmp/opsi-linux-client-agent/files/postinst.d/postinst.env\n\
    - echo "CLIENT_KEY=#@client_key*#" >> /target/var/tmp/opsi-linux-client-agent/files/postinst.d/postinst.env\n\
    - echo "SERVICE_ADDRESS=https://#@depotAddress*#:4447" >> /target/var/tmp/opsi-linux-client-agent/files/postinst.d/postinst.env\n\
    - curtin in-target --target=/target -- wget -r -nH --no-check-certificate --cut-dirs=3 --no-parent --http-user #@client_id*# --http-passwd #@client_key*# --reject="index.html*" https://#@depotAddress*#:4447/depot/#@productid*#/opsi/postinst.sh -P /var/tmp/opsi-linux-client-agent/files/\n\
    - curtin in-target --target=/target -- chmod +x /var/tmp/opsi-linux-client-agent/files/postinst.sh\n\
    - curtin in-target --target=/target -- wget -r -nH --no-check-certificate --cut-dirs=2 --no-parent --http-user #@client_id*# --http-passwd #@client_key*# --reject="index.html*" https://#@depotAddress*#:4447/depot/#@productid*#/olca-install.service -P /etc/systemd/system/\n\
    - curtin in-target --target=/target -- sed -i s/service_ip_address/#@depotAddress*#/g /etc/systemd/system/olca-install.service\n\
    - curtin in-target --target=/target -- systemctl enable olca-install.service'
if not username_from_property:
    userdel = '\n\
    - echo "#!/bin/bash" >> /target/var/tmp/opsi-linux-client-agent/files/postinst.d/10_userdel.sh\n\
    - echo "userdel -r #@username*#" >> /target/var/tmp/opsi-linux-client-agent/files/postinst.d/10_userdel.sh\n\
    - chmod +x /target/var/tmp/opsi-linux-client-agent/files/postinst.d/10_userdel.sh'
    opsi_client_agent_installation += userdel
if desktop_package and encrypt_password:
    desktop_show_enc_password = '\n\
    - sed -i \'s/GRUB_CMDLINE_LINUX_DEFAULT=\"\"/GRUB_CMDLINE_LINUX_DEFAULT=\"quiet splash\"/g\' /target/etc/default/grub \n\
    - curtin in-target --target=/target -- update-grub'
    opsi_client_agent_installation += desktop_show_enc_password

install_unattended = True
if (str(productProperties.get('install_unattended', 'true')).lower() == 'false'):
    install_unattended = False

installation_method = productProperties.get('installation_method', 'kexec')
if inSecureboot:
    installation_method = 'reboot'
    logger.notice("We are running in SecureBoot mode, enforcing reboot installation_method")

locale = productProperties.get('locale','de_DE')
language = locale.split('_')[0]
country = locale.split('_')[1]

online_repository = productProperties.get('online_repository', 'http://de.archive.ubuntu.com/ubuntu')

myarchitecture = getArchitecture()
if myarchitecture == 'amd64' and 'opsi' in online_repository:
    online_repository = online_repository + '/amd64'
if myarchitecture == 'i386' and 'opsi' in online_repository:
    online_repository = online_repository + '/i386'

logger.notice(u"splitting online_repository ...")
online_repository_protocol=online_repository.split('://')[0]
online_repo_adr = online_repository.split('://')[1]
online_repository_base=online_repo_adr.split('/')[0]
# get the rest: split has num argument
online_repository_dir='/'+online_repo_adr.split('/',1)[1]

partition_disk = productProperties.get('partition_disk', 'first')
partition_method = productProperties.get('partition_method', 'lvm')
encrypt_password = productProperties.get('encrypt_password', '')
if encrypt_password:
    logger.notice('encrypt_password is set, switching partition_method to lvm')
    partition_method = 'lvm'
partition_recipe = productProperties.get('partition_recipe', 'atomic')
if partition_recipe == 'customSwap':
    partition_method = 'regular'
    swapSizeProp = productProperties.get('swap_size', '')
    if swapSizeProp:
        removedUnits = removeUnit(swapSizeProp)
        swapSize = int(removedUnits)/1048576
        logger.notice(u"swap size: %s" % swapSize)
    else:
        meminfo = open('/proc/meminfo').read()
        matched = re.search(r'^MemTotal:\s+(\d+)', meminfo)
        if matched:
            memory = int(matched.groups()[0])
            swapSize = int(memory)/1024
            logger.notice(u"memory: %s" % swapSize)


if inUefi:
    partition_recipe = 'amd64/efi'
#language_packs = productProperties.get('language_packs', 'standard')
upgrade_select = productProperties.get('upgrade_select', 'none')


my_broadcast = getBroadcastAddress(ipaddress, netmask)
my_network = getNetworkAddress(ipaddress, netmask)

if (proxy == ''):
    do_proxy = 'false'
    proxy_addr = ''
else:
    do_proxy = 'true'
    proxy_addr = proxy
    if proxy_addr.startswith('https:'):
        execute('export HTTPS_PROXY="%s"' % proxy_addr)
    else:
        execute('export HTTP_PROXY="%s"' % proxy_addr)

opsi_client_agent_install_script = ''

# get machine architecture
scriptMessageSubject.setMessage(u"Installing 64 Bit system....")
logger.notice(u"Installing 64 Bit system....")
myarchitectureBits = '64'

scriptMessageSubject.setMessage("Create temporary place on disk.... ")
logger.notice(u"Create temporary place on disk.... ")
if isOpsi43:
    show_message("Preparing disk", 20)
#prepareDisk()
target = '/mnt/hd'
systemPartitionNumber = 1
if inUefi:
    systemPartitionNumber = 3
    if installation_method == 'reboot':
        grubPartitionNumber = 3
        systemPartitionNumber +=1
        grubPartitionSizeM = 2000
        grubPartitionLabel = 'INSTALLER'
systemPartitionSizeM = 8000
systemPartitionLabel  = 'LINUX'
blockAlignment = True

logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
lvmClearVolumeGroups()
execute('sfdisk -l')

# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
# use first disk
disk = disks[0]
if partition_disk == "first":
    partition_disk = disk.device
disk_serial = ''
try:
    disk_serial = execute(f'udevadm info --query=all --name={partition_disk} | grep ID_SERIAL= | cut -d= -f2-')[0].replace(' ','_')
except IndexError:
    logger.warning("reading the hard drive serial failed.")
    logger.warning("Installing OS on largest disk if multiple disks are installed.")

if useGpt:
    logger.notice(u"Clear partition table of %s" % (disk.device))
    execute('/usr/sbin/sgdisk --zap-all %s' % (disk.device))
    execute('/usr/sbin/sgdisk --clear -g -p %s' % (disk.device))
    if inUefi:
        currentbootentry = uefiGetActBootEntry()
        if currentbootentry:
            uefiNetbootLabel = currentbootentry['label']
            logger.notice('Current (net) boot entry: %s' % uefiNetbootLabel)
            backend.config_createUnicode("clientConfig.uefiNetbootLabel","clientspecific uefi firmware netboot entry label")
            backend.configState_create("clientConfig.uefiNetbootLabel", clientId, [ uefiNetbootLabel ] )
        else:
            logger.notice('No usable current (net) boot entry found')

        deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
        deleteAllEntriesFromEFIBootmgr('Windows Boot Manager')
        deleteAllEntriesFromEFIBootmgr('ubuntu')
        deleteAllEntriesFromEFIBootmgr('rocky')
        deleteAllEntriesFromEFIBootmgr('debian')
        deleteAllEntriesFromEFIBootmgr('installer')
        deleteAllEntriesFromEFIBootmgr('opsibootimage')
        logger.notice(u"Create EFI system partition on %s1" % (disk.device))
        execute('/usr/sbin/sgdisk -g --new 1::100M -t 1:ef00 -c 1:\"EFI system partition\" --partition-guid=C12A7328-F81F-11D2-BA4B-00A0C93EC93B -p %s' % (disk.device))
        execute('/bin/sleep 2')
        execute('/usr/sbin/sgdisk -p %s' % (disk.device))
        logger.notice(u"Make file system on EFI system partition on %s1" % (disk.device))
        normalDiskDevices = ['/dev/sda', '/dev/vda', '/dev/sdb']
        if any(normalDisk in disk.device for normalDisk in normalDiskDevices):
            execute('/sbin/mkfs.vfat %s1' % (disk.device))
        else:
            execute('/sbin/mkfs.vfat %sp1' % (disk.device))
        logger.notice(u"Create MS reserved partition on %s2" % (disk.device))
        execute('/usr/sbin/sgdisk -g --new 2::+128M -t 2:0c01 -c 2:\"MS reserved\" --partition-guid=E3C9E316-0B5C-4DB8-817D-F92DF00215AE -p %s' % (disk.device))
else:
    disk.deletePartitionTable()

if installation_method == 'reboot':
    if inUefi:
        grubTarget = '/mnt/grub/'
        checkCreateDir(grubTarget)
        execute('/usr/sbin/sgdisk --info=%d %s' % (grubPartitionNumber, disk.device))
        createPartitionEx("installer", grubPartitionNumber, grubPartitionLabel, 0, grubPartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "63" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "fat32", gptFileSystem="0700")
        execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 --attributes %d:clear:0 -p %s' % (grubPartitionNumber, grubPartitionNumber, grubPartitionNumber, disk.device))
        testPartitionEx(disk.device, grubPartitionNumber, useGpt)
        checkClearPartition("none", disk.device, grubPartitionNumber)
        createFilesystemEx("bootimage", grubPartitionNumber, "bootimage", "fat32", disk.device , useGpt)
        checkMount(grubTarget, disk.device, grubPartitionNumber, "vfat")
    else:
        logger.warning(u"reboot installation method only works using UEFI. Reverting back to kexec method")
        installation_method = "kexec"

fdiskexitcode = os.system('%s %s >> /dev/null 2>&1' % (which('partprobe'), disk.device))
createPartitionEx("system", systemPartitionNumber, systemPartitionLabel, 0, systemPartitionSizeM, gptAutoAppend = True, boot = False, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ext4", gptFileSystem="8300")
createFilesystemEx("system", systemPartitionNumber, systemPartitionLabel, "ext4", disk.device, useGpt)
systemPartitionDevicePath = getPartitionDevicePath(disk.device, systemPartitionNumber)

# Partition table deleted => delete all netboot product states
deletePocs = []
for poc in backend.productOnClient_getObjects(clientId = clientId):
    if ((poc.productType == 'NetbootProduct') and (poc.productId != productId)) or \
       ((poc.productType == 'LocalbootProduct') and (poc.installationStatus == 'not_installed') and (poc.actionRequest == 'none')):
        deletePocs.append(poc)
backend.productOnClient_deleteObjects(deletePocs)

if isOpsi43:
    show_message("Mounting partitions and copying data", 40)
# Mount temporary partition
checkCreateDir(target)
checkClearMountpoint(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkClearPartition(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkMount(target, disk.device, systemPartitionNumber, "ext4", systemPartitionDevicePath)
checkCreateDir(target+'/tmp')

scriptMessageSubject.setMessage("Getting iso.... ")
logger.notice(u"Getting iso.... ")

depotAddress = depotUrl.split('/')[2].replace(':4447','')

#isofilename = subprocess.Popen(f"wget -d -r -np -N --spider -e robots=off --no-check-certificate   https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/iso/   2>&1 | grep ' -> ' | grep -Ev '\/\?C=' | sed 's/.* -> //' | grep '.*\.iso$' | rev | cut -d '/' -f 1 | rev", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0].decode("utf-8").rstrip()
try:
    isofilename = execute(f"wget -d -r -np -N --spider -e robots=off --no-check-certificate   https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/iso/   2>&1 | grep ' -> ' | grep -Ev '\/\?C=' | sed 's/.* -> //' | grep '.*\.iso$' | rev | cut -d '/' -f 1 | rev")[0]
except IndexError:
    raise Exception('No iso in iso directory found')
logger.notice(u"Will use iso file: %s " % isofilename)

scriptMessageSubject.setMessage("Getting kernel and initrd.... ")
logger.notice(u"Getting kernel and initrd.... ")

execute(f'mkdir {target}/iso')
depot.copy(f'/{productId}/iso/{isofilename}', f'{target}/iso/')
execute('mkdir -p /mnt/iso/')
execute(f'mount {target}/iso/{isofilename} /mnt/iso/')
execute(f'cp /mnt/iso/casper/vmlinuz {target}/tmp/')
execute(f'cp /mnt/iso/casper/initrd {target}/tmp/initrd.raw')

block1Output = execute(f"dd if={target}/tmp/initrd.raw skip=0 | cpio -it")
for line in block1Output:
    match = re.search("^[-+]?[0-9]+", line)
    if match:
        block1 = forceInt(match.group(0))
logger.notice(f'initrd: skipping first {block1} blocks')
block2Output = execute(f"dd if={target}/tmp/initrd.raw skip={block1} | cpio -it")
for line in block2Output:
    match = re.search("^[-+]?[0-9]+", line)
    if match:
        block2 = forceInt(match.group(0))
kernel_block_sum = int(block1) + int(block2)
block3Output = execute(f"dd if={target}/tmp/initrd.raw skip={kernel_block_sum} | cpio -it")
for line in block3Output:
    match = re.search("^[-+]?[0-9]+", line)
    if match:
        block3 = forceInt(match.group(0))
fs_block_sum = int(block1) + int(block2) + int(block3)
execute(f'dd if={target}/tmp/initrd.raw skip={kernel_block_sum} of={target}/tmp/initrd.fs')
execute(f'dd if={target}/tmp/initrd.raw skip={fs_block_sum} of={target}/tmp/initrd.zst')
execute(f'unzstd {target}/tmp/initrd.zst')

scriptMessageSubject.setMessage("Uncompressing initrd.... ")
logger.notice(u"Uncompressing initrd.... ")
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass

scriptMessageSubject.setMessage("Patching initrd.... ")
logger.notice(u"Mounting initrd ....")
execute("mkdir -p %s/tmp/mnt" % (target))
execute("cd %s/tmp/mnt ; cpio -idm -H newc < ../initrd" % (target))
execute("cd %s/tmp/mnt ; cpio -idm -H newc < ../initrd.fs" % (target))

encrypted_lvm_installation = ""
logger.notice(u"Copy profile for patch....")
preseed = productProperties.get('autoinstall', 'auto')
if preseed == 'auto':
    depot.copy(SCRIPT_PATH+f'/{productId}.yml', target+'/tmp/installation_profile')
    if encrypt_password:
        encrypted_lvm_installation = "sizing-policy: all\n      password: #@encrypt_password*#"
else:
    logger.notice(u"Custom autoinstall (%s) found and will be used" % preseed)
    depot.copy(SCRIPT_PATH+'/custom/%s' % preseed, target+'/tmp/installation_profile')
logger.notice(u"Copy /etc/default/keyboard for patch....")
depot.copy(SCRIPT_PATH+'/keyboard', target+'/tmp/keyboard')


localUbuntuRepo = False
if 'archive.ubuntu.com/ubuntu' not in online_repository:
    localUbuntuRepo = True

logger.notice(u"Copy SYSCONF_INI....")
# Copy sysconf.ini to opsi-client-agent config dir
copy(SYSCONF_INI , target +'/tmp/')

if partition_recipe == 'customSwap':
    diskSizeM = (int(disk.size)/(1024*1024))
    rootSize = diskSizeM-swapSize-500

if isOpsi43:
    show_message("patching files", 90)
logger.notice(u"Prepare patch input....")
with open(PATCHA_IN, "a") as myfile:
    myfile.write('networkaddress=%s\n' % my_network)
    myfile.write('broadcastaddress=%s\n' % my_broadcast)
    myfile.write('nameserveraddress=%s\n' % nameserver)
    myfile.write('do_opsi_client=%s\n' % do_opsi_client)
    myfile.write('username=%s\n' % username)
    myfile.write('userpassenc=%s\n' % userpassenc)
    myfile.write('rootpassenc=%s\n' % rootpassenc)
    myfile.write('do_proxy=%s\n' % do_proxy)
    myfile.write('proxy_addr=%s\n' % proxy_addr)
    myfile.write('network_device=%s\n' % device)
    myfile.write('online_repository=%s\n' % online_repository)
    myfile.write('online_repository_protocol=%s\n' % online_repository_protocol)
    myfile.write('online_repository_base=%s\n' % online_repository_base)
    myfile.write('online_repository_dir=%s\n' % online_repository_dir)
    myfile.write('opsi_client_agent_install_script=%s\n' % opsi_client_agent_install_script)
    myfile.write('partition_disk=%s\n' % partition_disk)
    myfile.write('disk_serial=%s\n' % disk_serial)
    myfile.write('encrypt_password=%s\n' % encrypt_password)
    myfile.write('partition_method=%s\n' % partition_method)
    myfile.write('partition_recipe=%s\n' % partition_recipe)
    myfile.write('desktop_package=%s\n' % desktop_package)
    myfile.write('language=%s\n' % country.lower())
    myfile.write('country=%s\n' % country)
    myfile.write('locale=%s\n' % locale)
    myfile.write('upgrade_select=%s\n' % (upgrade_select))
    myfile.write('service_ip_address=%s\n' % (nextserver.replace("/rpc", "").rstrip("/")))
    myfile.write('depotAddress=%s\n' % (nextserver.replace("https://", "").replace(":4447","").replace("rpc", "").replace("/", "")))

logger.notice(u"Here comes the PATCHA_IN (%s):" % PATCHA_IN)
execute('cat %s' % (PATCHA_IN))

logger.notice(u"Patching installation_profile")
with open(target + '/tmp/installation_profile', 'r') as inputxml:
    filedata = inputxml.read()
    filedata = filedata.replace('#@opsi_client_agent_installation*#', opsi_client_agent_installation)
    filedata = filedata.replace('#@encrypted_lvm_installation*#', encrypted_lvm_installation)
    with open(target + '/tmp/installation_profile', 'w') as outputxml:
        outputxml.write(filedata)

if not disk_serial:
    execute(f"sed -i '/match:/d' {target}/tmp/installation_profile")
    execute(f"sed -i '/disk_serial/d' {target}/tmp/installation_profile")

if not desktop_package:
    execute(f"sed -i '/desktop_package/d' {target}/tmp/installation_profile")

for patchfile in ( target + '/tmp/installation_profile' , target+'/tmp/keyboard'):
    execute('/usr/local/bin/patcha %s' % patchfile)
    execute('/usr/local/bin/patcha -f %s %s' % (PATCHA_IN, patchfile))

######### enable root login on console if password is set
rootPassword = productProperties.get('root_password','')

######## how to install packages
if install_unattended:
    logger.notice(u"Copy profile file to initrd...")
    execute(f"mkdir {target}/tmp/mnt/autoinstall/")
    copy(target+'/tmp/installation_profile', target+'/tmp/mnt/autoinstall/user-data')
    logger.notice(u"Here comes the installation profile:")
    execute('cat %s/tmp/mnt/autoinstall/user-data' % (target))
    execute('touch %s/tmp/mnt/autoinstall/meta-data' % (target))
    execute('touch %s/tmp/mnt/autoinstall/vendor-data' % (target))
    with open(f"{target}/tmp/mnt/conf/conf.d/autoinstall.conf", "a") as aiconf:
        aiconf.write("echo 'cp -r /autoinstall /run/\n' >> /scripts/init-top/ORDER")

logger.notice(u"Copy opsi-client-agent file to initrd...")
execute('mkdir -p %s/tmp/mnt/opsitmp/' % (target))
copy(target+'/tmp/sysconf.ini', target+'/tmp/mnt/opsitmp/')
copy(target+'/tmp/keyboard', target+'/tmp/mnt/opsitmp/')
copy(PATCHA_IN, target+'/tmp/mnt/opsitmp/')

execute('mkdir -p %s/tmp/mnt/opsitmp/' % (target))
copy(PATCHA_IN, target+'/tmp/mnt/opsitmp/')

execute(f"rm -rf {target}/tmp/mnt/usr/bin/wget")
depot.copy(SCRIPT_PATH + '/installfiles%s/wget' % myarchitectureBits, target+'/tmp/mnt/usr/bin/')
execute(f"chmod +x {target}/tmp/mnt/usr/bin/wget")

execute(f"wget --no-check-certificate -P {target}/tmp/mnt/etc/ssl/certs/ https://{fqdn}:{pckey}@{depotAddress}:4447/ssl/opsi-ca-cert.pem")
execute(f"cat {target}/tmp/mnt/etc/ssl/certs/opsi-ca-cert.pem >> {target}/tmp/mnt/usr/lib/ssl/cert.pem")

if (str(productProperties.get('ignore_certificate_errors', 'false')).lower() == 'false'):
    execute(f"sed -i 's_wget_/usr/bin/wget_g' {target}/tmp/mnt/scripts/casper")
else:
    execute(f"sed -i 's_wget_/usr/bin/wget --no-check-certificate_g' {target}/tmp/mnt/scripts/casper")

execute("rm %s/tmp/initrd" % (target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("cd %s/tmp/mnt ; find . | cpio -o -H newc > %s/tmp/initrd" % (target, target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass

if installation_method == 'reboot':
    logger.notice("Copying files to partition %d" % grubPartitionNumber)
    depot.copy(f'/{productId}/installfiles64/*', grubTarget)
    subiquityLine = f' autoinstall ds="nocloud;s=file://run/autoinstall/" url=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/iso/{isofilename} cloud-config-url=/dev/null ip=dhcp locale={locale}'
    newLines = []
    with open('%s/grub/grub.cfg' % (grubTarget), 'r+') as grubConfig:
        for line in grubConfig:
            logger.notice(f"GRUB LINE: {line}")
            if line.strip().startswith('linux'):
                newLines.append('%s %s\n' % (line.rstrip(), subiquityLine.rstrip()))
                continue
            newLines.append(line)

    with open('%s/grub/grub.cfg' % (grubTarget), 'w') as grubConfig:
        grubConfig.writelines(newLines)
    execute(f'sed -i s_preseed/file=/preseed.cfg__g {grubTarget}/grub/grub.cfg')
    execute(f'sed -i s_/linux_/vmlinuz_g {grubTarget}/grub/grub.cfg')
    execute(f'sed -i s/DEBCONF_DEBUG=5//g {grubTarget}/grub/grub.cfg')
    

    execute('/bin/efibootmgr -c --disk %s --part %s -L installer --loader \\\\bootx64.efi' % (disk.device, grubPartitionNumber))
    setFirstUefiBoot('installer')
    setNextUefiBoot('installer')
    execute('/bin/efibootmgr -v')
    execute(f"cp {target}/tmp/initrd* {grubTarget}")
    execute(f"cp {target}/tmp/vmlinuz {grubTarget}")
    execute(f"cp /mnt/iso/EFI/boot/bootx64.efi {grubTarget}")
    execute(f"cp /mnt/iso/EFI/boot/grubx64.efi {grubTarget}")
    execute(f"cp -r /mnt/iso/boot/ {grubTarget}")
    execute(f"cp -r  {grubTarget}/grub/grub.cfg {grubTarget}/boot/grub/grub.cfg")
else:
    scriptMessageSubject.setMessage("Prepare loading kernel.... ")
    logger.notice(u"Prepare loading kernel ....")
    kernel_file = target+'/tmp/vmlinuz'
    append_line = f'autoinstall ds=nocloud;s=file://run/autoinstall/ url=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/iso/{isofilename} cloud-config-url=/dev/null ip=dhcp  ---'
    inird_file = target+'/tmp/initrd'
    execute("%s --load %s  --append='%s' --initrd=%s" % (which('kexec'), kernel_file, append_line, inird_file))

logger.notice(u"handle_setupAfterInstall ....")
handle_setupAfterInstall()

# Set product installation status
productOnClient.setInstallationStatus('unknown')
#productOnClient.setActionProgress('installing')
productOnClient.setActionResult('')
productOnClient.setLastAction('setup')

#scriptMessageSubject.setMessage("Writing log to service.... ")
try:
    # Write log to service
    logger.notice(u"Writing log to service")

    log = codecs.open(LOG_FILE, 'r', 'utf-8', 'replace')
    data = log.read()
    log.close()

    logger.setFileLevel(LOG_ERROR)
    if backend:
        try:
            try:
                ui.getMessageBox().addText(_(u"Writing log to service\n"))
            except:
                pass
            backend.log_write(logType = 'bootimage', data = data, objectId = clientId, append = False)
        except Exception as e:
            logger.error(u"Failed to write log to service: %s" % e)

except Exception as e:
    logger.logException(e)

if installation_method == 'reboot':
    scriptMessageSubject.setMessage("Rebooting machine ")
    logger.notice(u"Rebooting machine")
    reboot()
else:
    scriptMessageSubject.setMessage("Execute new kernel.... ")
    logger.notice(u"Execute new kernel ....")
    execute("%s -e" % (which('kexec')))
